"""Utils for HDF5 file handling"""
# Installed
import h5py as h5


def h5dump(f: h5.File or h5.Group, include_attrs: bool = True, stdout: bool = False):
    """Prints the contents of an HDF5 object.

    Parameters
    ----------
    f: h5.File or h5.Group
        File, Group object from which to start inspecting.
    include_attrs: bool, Optional
        Default True.
    stdout: bool, Optional
        Default False. If True, prints to stdout as the object tree is traversed.

    Returns
    -------
    : str
        Concatenated string of HDF5 contents
    """
    if not isinstance(f, (h5.File, h5.Group)):
        raise ValueError(f"Invalid input to h5dump. H5 object f must be a File or Group. Got {type(f)}")
    srep = []

    def _print(name, obj):
        if isinstance(obj, h5.Group):
            s = f"Group:{obj.name} ({len(obj)} members, {len(obj.attrs) if obj.attrs else 0} attributes)"
            srep.append(s + '\n')
        elif isinstance(obj, h5.Dataset):
            s = (f"Dataset:{obj.name} "
                 f"(shape={obj.shape}, type={obj.dtype}, {len(obj.attrs) if obj.attrs else 0} attributes)")
            srep.append(s + '\n')
        elif isinstance(obj, h5.Datatype):
            s = f"Datatype:{obj.name} {obj}"
            srep.append(s + '\n')
        else:
            raise ValueError(f"Unrecognized object discovered in h5dump, of type {type(obj)}.")

        if stdout:
            print(s)

        if include_attrs and obj.attrs:
            for key, val in obj.attrs.items():
                s = f"    @ {key} = {val}"
                srep.append(s + '\n')
                if stdout:
                    print(s)

    top_obj_name = f.name if f.name == '/' else f.name[1:]  # Creates a name similar to that passed by visititems
    _print(top_obj_name, f)
    f.visititems(_print)
    return "".join(srep)



"""Logging utilities"""
# Standard
import logging
import logging.config
import logging.handlers
import yaml
# Installed
from cloudpathlib import AnyPath
import watchtower
# Local
from libera_utils.config import config
from libera_utils.io.smart_open import smart_open

logger = logging.getLogger(__name__)


def configure_static_logging(config_file: AnyPath or str):
    """Configure logging based on a static logging configuration yaml file.

    The yaml is interpreted as a dict configuration. There is no ability to customize this logging
    configuration at runtime.

    Parameters
    ----------
    config_file : AnyPath or str
        Location of config file.

    See Also
    --------
    configure_task_logging : Runtime modifiable logging configuration.
    """
    with smart_open(config_file) as log_config:
        config_yml = log_config.read()
        config_dict = yaml.safe_load(config_yml)
    logging.config.dictConfig(config_dict)
    logger.info(f"Logging configured statically according to {config_file}.")


def configure_task_logging(task_id: str, app_package_name: str, console_log_level: str or int = None):
    """Configure logging based on runtime environment variables.

    Variables that control logging are LIBERA_LOG_DIR, LIBERA_CONSOLE_LOG_LEVEL, and LIBERA_LOG_GROUP. If these
    variables are unset, only INFO level console logging will be enabled.

    Parameters
    ----------
    task_id : str
        Unique identifier by which to name the log file and cloudwatch log stream.
    app_package_name : str
        This is the name of the top level package for which you want to instantiate logging. For example, if you are
        working on an application package called `my_app` and using module level logging, all your loggers will be
        named like `my_app.module_name.submodule_name`. We use this string to set the logging level of all loggers
        that inherit from the `my_app` logger (logger inheritance in python is expressed in dot notation). So by
        specifying `my_app` as the app_package_name, all your app logger handlers will log at your specified levels
        but all library loggers (e.g. those not inheriting from `my_app` logger) will only log at INFO level.
        This reduces debug spam from library loggers significantly, especially boto3.
    console_log_level : str or int, Optional
        Override environment variable log level configuration.

    See Also
    --------
    configure_static_logging : Static logging configuration based on yaml file.
    """
    def _str_bool(s: str):
        """Examines an environment variable string to determine if it is truthy or falsy"""
        if not bool(s):
            return False
        if s.lower() in ("false", "0", "none", "null"):
            return False
        return True

    handlers = {}
    setup_messages = []
    try:  # Establish console log level from config
        if not console_log_level:
            console_log_level = config.get("LIBERA_CONSOLE_LOG_LEVEL")
            if _str_bool(console_log_level):
                console_log_level = console_log_level.upper()

        if console_log_level:
            console_handler = {
                "class": "logging.StreamHandler",
                "formatter": "plaintext",
                "level": console_log_level.upper(),
                "stream": "ext://sys.stdout"
            }
            handlers.update(console=console_handler)
            setup_messages.append(f"Console logging configured at level {console_log_level}.")
    except KeyError:
        pass

    try:  # Establish log directory from config
        log_dir = config.get("LIBERA_LOG_DIR")
        if _str_bool(log_dir):
            log_filepath = AnyPath(log_dir) / f"{task_id}.log"
            logfile_handler = {
                "class": "logging.handlers.RotatingFileHandler",
                "formatter": "plaintext",
                "level": "DEBUG",
                "filename": str(log_filepath),
                "maxBytes": 1000000,
                "backupCount": 3
            }
            handlers.update(logfile=logfile_handler)
            setup_messages.append(f"File logging configured to log to {log_filepath}.")
    except KeyError:
        pass

    try:  # Establish cloudwatch log group from config
        log_group = config.get("LIBERA_LOG_GROUP")
        if _str_bool(log_group):
            watchtower_handler = {
                "class": "watchtower.CloudWatchLogHandler",
                "formatter": "json",
                "level": "DEBUG",
                "log_group_name": log_group,
                "log_stream_name": task_id,
                "send_interval": 10,
                "create_log_group": False
            }
            handlers.update(watchtower=watchtower_handler)
            setup_messages.append(f"Cloudwatch logging configured for log-group/log-stream: {log_group}/{task_id}.")
    except KeyError:
        pass

    config_dict = {
        "version": 1,
        "disable_existing_loggers": False,
        "formatters": {
            "json": {
                "format": '{"time": "%(asctime)s", level": "%(levelname)s", "module": "%(filename)s", '
                          '"function": "%(funcName)s", "line": %(lineno)d, "message": "%(message)s"}',
            },
            "plaintext": {
                "format": "%(asctime)s %(levelname)-9.9s [%(filename)s:%(lineno)d in %(funcName)s()]: %(message)s"
            }
        },
        "handlers": handlers,
        "root": {
            "level": "INFO",
            "propagate": True,
            "handlers": list(handlers.keys())
        },
        "loggers": {
            app_package_name: {
                "level": "DEBUG",
                "handlers": []
            }
        }
    }

    logging.config.dictConfig(config_dict)
    for message in setup_messages:
        logger.info(message)


def flush_cloudwatch_logs():
    """Force flush of all cloudwatch logging handlers. For example at the end of a process just before it is killed.

    Returns
    -------
    None
    """
    root_logger = logging.getLogger()
    for handler in root_logger.handlers:
        if isinstance(handler, watchtower.CloudWatchLogHandler):
            handler.flush()



"""Module containing code to manage local file caching"""
# Standard
import os
import sys
from importlib.metadata import version
from pathlib import Path


def get_local_cache_dir():
    """Determine where to cache files based on the system and installed package version.

    Returns
    -------
    : Path
        Path to the cache directory for this version of this package on the current system
    """
    system = sys.platform
    package_name = __name__.split('.', 1)[0]
    package_version = version(package_name)
    if system == 'darwin':
        path = Path('~/Library/Caches').expanduser()
        if package_name:
            path = path / package_name
    elif system.startswith('linux'):
        if os.getenv('XDG_CACHE_HOME'):
            path = Path(os.getenv('XDG_CACHE_HOME'))
        else:
            path = Path('~/.cache').expanduser()
        if package_name:
            path = path / package_name
    else:
        raise NotImplementedError("Only MacOS (darwin) and Linux (linux) platforms are currently supported. "
                                  f"Unsupported platform appears to be {system}")
    if package_name and package_version:
        path = path / package_version
    return path


def empty_local_cache_dir():
    """Remove all cached files in the local cache.

    Returns
    -------
    list
        List of removed files
    """
    removed_files = []
    local_cache = get_local_cache_dir()
    for cached_file in local_cache.glob("*"):
        os.remove(cached_file)
        removed_files.append(cached_file)
    return removed_files



"""Module for the Libera SDC utilities CLI

libera-utils
    make-kernel
        jpss-spk
        jpss-ck
        azel-ck
"""
# Standard
import argparse
# Local
from libera_utils import kernel_maker
from libera_utils.version import version as libera_utils_version


def main(cli_args: list = None):
    """Main CLI entrypoint that runs the function inferred from the specified subcommand"""
    args = parse_cli_args(cli_args)
    args.func(args)


def print_version_info(*args):
    """Print CLI version information"""
    print(f"Libera SDC utilities CLI\n\tVersion {libera_utils_version()}"
          f"\n\tCopyright 2022 University of Colorado\n\tReleased under BSD3 license")


def parse_cli_args(cli_args: list):
    """Parse CLI arguments

    Parameters
    ----------
    cli_args : list
        List of CLI arguments to parse

    Returns
    -------
    Namespace
        Parsed arguments in a Namespace object
    """
    parser = argparse.ArgumentParser(prog="libera-utils", description="Libera SDC utilities CLI")
    parser.add_argument("--version",
                        action='store_const', dest='func', const=print_version_info,
                        help="print current version of the CLI")

    subparsers = parser.add_subparsers(description="sub-commands for libera-utils CLI")

    # make-kernel
    make_kernel_parser = subparsers.add_parser('make-kernel',
                                               help='generate SPICE kernel from telemetry data')

    make_kernel_subparsers = make_kernel_parser.add_subparsers(description="sub-commands for make-kernel sub-command")

    # make-kernel jpss-spk
    jpss_spk_parser = make_kernel_subparsers.add_parser('jpss-spk', help="generate JPSS SPK kernel from telemetry")
    jpss_spk_parser.set_defaults(func=kernel_maker.make_jpss_spk)
    jpss_spk_parser.add_argument('packet_data_filepaths', nargs='+', type=str,
                                 help="paths to L0 packet files")
    jpss_spk_parser.add_argument('--outdir', '-o', type=str,
                                 required=True,
                                 help="output directory for generated SPK")
    jpss_spk_parser.add_argument('--overwrite', action='store_true',
                                 help="force overwriting an existing kernel if it exists")
    jpss_spk_parser.add_argument('-v', '--verbose', action='store_true',
                                 help="set DEBUG level logging output (otherwise set by LIBERA_CONSOLE_LOG_LEVEL)")

    # make-kernel jpss-ck
    jpss_ck_parser = make_kernel_subparsers.add_parser('jpss-ck', help="generate JPSS CK kernel from telemetry")
    jpss_ck_parser.set_defaults(func=kernel_maker.make_jpss_ck)
    jpss_ck_parser.add_argument('packet_data_filepaths', nargs='+', type=str,
                                help="paths to L0 packet files")
    jpss_ck_parser.add_argument('--outdir', '-o', type=str,
                                required=True,
                                help="output directory for generated CK")
    jpss_ck_parser.add_argument('--overwrite', action='store_true',
                                help="force overwriting an existing kernel if it exists")
    jpss_ck_parser.add_argument('-v', '--verbose', action='store_true',
                                help="set DEBUG level logging output (otherwise set by LIBSDP_STREAM_LOG_LEVEL)")

    # make-kernel azel-ck
    azel_ck_parser = make_kernel_subparsers.add_parser('azel-ck', help="generate Libera Az-El CK kernel from telemetry")
    azel_ck_parser.set_defaults(func=kernel_maker.make_azel_ck)
    azel_ck_parser.add_argument('packet_data_filepaths', nargs='+', type=str,
                                help="paths to L0 packet files")
    azel_ck_parser.add_argument('--outdir', '-o', type=str,
                                required=True,
                                help="output directory for generated CK")
    azel_ck_parser.add_argument('--overwrite', action='store_true',
                                help="force overwriting an existing kernel if it exists")
    azel_ck_parser.add_argument('-v', '--verbose', action='store_true',
                                help="set DEBUG level logging output (otherwise set by LIBSDP_STREAM_LOG_LEVEL)")

    parsed_args = parser.parse_args(cli_args)
    return parsed_args



"""Configuration reader. To modify the configuration, see file: emus_config.json"""
# Standard
import json
import logging
import os
from pathlib import Path
import re
import string
import sys
import warnings

logger = logging.getLogger(__name__)


class ConfigurationFormatter(string.Formatter):
    """Customize the string formatter to replace fields in a config string
    with values from the configuration dictionary. This will allow configuration
    parameters in the emus_config.json file to be based off of other configuration
    parameters by wrapping the configuration key in curly braces."""

    def get_value(self, key: str, *args, **kwargs):
        """Overrides the default get_value method in the python formatter. This will
        return the value from the emus configuration with the specified key."""
        return config.get(key)


class _ConfigurationCache:
    """Class that stores the JSON configuration and provides methods for accessing configuration information"""

    _json_config_filepath = Path(__file__).parent / 'data/config.json'

    # Should handle all floats
    FLOAT_REGEXP = re.compile(r'^[-+]?([0-9]+|[0-9]*\.[0-9]+)([eE][-+]?[0-9]+)?$')
    # Example "-12.34E+56"      # sign (-)
    #     integer (12)
    #           OR
    #             int/mantissa (12.34)
    #                            exponent (E+56)

    def __init__(self):
        logger.debug("Initializing configuration cache from config.json.")
        with open(self._json_config_filepath, encoding='utf_8') as config_file:
            self._cached_json_config = json.load(config_file)
            self._known_config_variables = set(self._cached_json_config)

    def _format_return_value(self, value: str):
        """Recursively formats the returned value, looking for config keys to substitute.

        Parameters
        ----------
        value : str
            String to format

        Returns
        -------
        str
        """
        if isinstance(value, str):
            formatter = ConfigurationFormatter()
            # if the string contains only a curly bracket keyword, return the value of get_config(keyword)
            key_iter = formatter.parse(value)
            text_and_key = next(key_iter)
            if text_and_key[0] == '' and not any(key_iter):
                return config.get(text_and_key[1])

            return formatter.format(value)

        if isinstance(value, list):
            return [self._format_return_value(x) for x in value]

        if isinstance(value, dict):
            return {k: self._format_return_value(v) for k, v in value.items()}

        return value

    def _parse_numeric_types(self, value: str):
        """Checks the final result of a config retrieval. If it is a string that can be interpreted as a float or int,
        parse it and return that.

        Parameters
        ----------
        value : any
            Final formatted value.

        Returns
        -------
        : str or float or int
        """
        if not isinstance(value, str):  # Indicates a non-string type came from the JSON config. Leave it as is.
            return value

        if self.FLOAT_REGEXP.match(value):
            f = float(value)
            if f.is_integer():
                return int(f)
            return f
        return value

    def force_reload(self):
        """Force reloading of the JSON config"""
        self.__init__()  # pylint: disable=C2801

    def get(self, key):
        """Retrieves a configuration value from either the cached JSON or from the environment

        Parameters
        ----------
        key : str
            Key for which to retrieve the configured value.

        Returns
        -------
        : any
            Resulting value
        """
        if (key != 'PKG_ROOT') and (key not in self._known_config_variables):
            warnings.warn(f"Configuration key {key} is not known to the config module. "
                          f"We will try to find it in the environment anyway but a default should be added to "
                          f"the config.json file.")

        if key is None:
            return self._cached_json_config

        if key == 'PKG_ROOT':  # Special case for root of installed package
            return str(Path(sys.modules[__name__.split('.', maxsplit=1)[0]].__file__).parent)

        # Checking the environment first allows easy override of any json config variable
        if os.getenv(key):
            result = os.getenv(key)
            return self._parse_numeric_types(self._format_return_value(result))

        if key in self._cached_json_config:
            result = self._cached_json_config[key]
            return self._parse_numeric_types(self._format_return_value(result))

        raise KeyError(f'Configuration variable {key} not found.')


config = _ConfigurationCache()
"""Convenience alias for :meth:`libera_utils.config._ConfigurationCache`"""



"""Module for file naming utilities"""
# Standard
from abc import ABC, abstractmethod
from datetime import datetime
from enum import Enum
import re
from types import SimpleNamespace
from pathlib import Path
# Installed
from cloudpathlib import AnyPath, CloudPath
# Local
from libera_utils.time import PRINTABLE_TS_FORMAT


SPK_REGEX = re.compile(r"^libera_(?P<spk_object>jpss)"
                       r"_(?P<utc_start>[0-9]{8}(?:t[0-9]{6})?)"
                       r"_(?P<utc_end>[0-9]{8}(?:t[0-9]{6})?)"
                       r"\.bsp$")

CK_REGEX = re.compile(r"^libera_(?P<ck_object>jpss|azrot|elscan)"
                      r"_(?P<utc_start>[0-9]{8}(?:t[0-9]{6})?)"
                      r"_(?P<utc_end>[0-9]{8}(?:t[0-9]{6})?)"
                      r"\.bc$")

LIBERA_PRODUCT_REGEX = re.compile(r"^libera"
                                  r"_(?P<instrument>cam|rad)"
                                  r"_(?P<level>l0|l1b|l2)"
                                  r"_(?P<utc_start>[0-9]{8}t[0-9]{6})"
                                  r"_(?P<utc_end>[0-9]{8}t[0-9]{6})"
                                  r"_(?P<version>vM[0-9]*m[0-9]*p[0-9]*)"
                                  r"_(?P<revision>r[0-9]{11})"
                                  r"\.(?P<extension>pkts|h5)$")

MANIFEST_FILE_REGEX = re.compile(r"^libera"
                                 r"_(?P<manifest_type>input|output)"
                                 r"_manifest"
                                 r"_(?P<created_time>[0-9]{8}(?:t[0-9]{6})?)"
                                 r"\.json")


class DataLevel(Enum):
    """Data product level"""
    L0 = "l0"
    L1B = 'l1b'
    L2 = 'l2'


class ManifestType(Enum):
    """Enumerated legal manifest type values"""
    INPUT = 'INPUT'
    input = INPUT
    OUTPUT = 'OUTPUT'
    output = OUTPUT


class AbstractValidFilename(ABC):
    """Composition of a CloudPath/Path instance with some methods to perform
    regex validation on filenames
    """
    _regex: re.Pattern
    _fmt: str

    def __init__(self, *args, **kwargs):
        self.path = AnyPath(*args, **kwargs)

    def __str__(self):
        return str(self.path)

    def __repr__(self):
        return f"{self.__class__.__name__}({str(self.path)})"

    def __eq__(self, other):
        if self.path == other.path and self.filename_parts == other.filename_parts:
            return True
        return False

    @property
    def path(self):
        """Property containing the file path"""
        return self._path

    @path.setter
    def path(self, new_path: str or Path or CloudPath):
        if isinstance(new_path, str):
            new_path = AnyPath(new_path)
        self.regex_match(new_path)  # validates against regex pattern
        self._path: CloudPath or Path = AnyPath(new_path)

    @property
    def filename_parts(self):
        """Property that contains a namespace of filename parts"""
        return self._parse_filename_parts()

    @classmethod
    def from_filename_parts(cls,
                            basepath: str or Path = None,
                            **parts):
        """Create instance from filename parts.

        The part arg names are named according to the regex for the file type.

        Parameters
        ----------
        basepath : str or Path, Optional
            Allows prepending a basepath for local filepaths. Does not work with
            cloud paths because there is no notion of a basepath in cloud storage.
        parts :
            Passed directly to _format_filename_parts

        Returns
        -------
        : cls
        """
        filename = cls._format_filename_parts(**parts)
        if basepath is not None:
            return cls(basepath, filename)
        return cls(filename)

    @classmethod
    @abstractmethod
    def _format_filename_parts(cls, **parts):
        """Format parts into a filename

        Note: When this is implemented by concrete classes, **parts becomes a set of explicitly named arguments"""
        raise NotImplementedError()

    @abstractmethod
    def _parse_filename_parts(self):
        """Parse the filename parts into objects from regex matched strings

        Returns
        -------
        : SimpleNamespace
            namespace object containing filename parts as parsed objects
        """
        d = self.regex_match(self.path)
        # Do stuff to parse the elements of d into a SimpleNamespace
        raise NotImplementedError()

    def regex_match(self, path: str or Path or CloudPath):
        """Parse and validate a given path against class-attribute defined regex

        Returns
        -------
        : dict
        """
        # AnyPath is polymorphic but self.path will always be a CloudPath or Path object with a name attribute.
        match = self._regex.match(path.name)  # pylint: disable=no-member
        if not match:
            raise ValueError(f"Proposed path {path} failed validation against regex pattern {self._regex}")
        return match.groupdict()


class ProductFilename(AbstractValidFilename):
    """Abstract base for filenaming and validation."""

    _regex = LIBERA_PRODUCT_REGEX
    _fmt = "libera_{instrument}_{level}_{utc_start}_{utc_end}_{version}_{revision}.{extension}"

    @classmethod
    def _format_filename_parts(cls,  # pylint: disable=arguments-differ
                               instrument: str,
                               level: DataLevel,
                               utc_start: datetime,
                               utc_end: datetime,
                               version: str,
                               revision: str,
                               extension: str):
        """Construct a path from filename parts

        Parameters
        ----------
        instrument : str
        level : DataLevel
        utc_start : datetime
            First timestamp in the SPK
        utc_end : datetime
            Last timestamp in the SPK
        version : str
            Software version that the file was created with. Corresponds to the algorithm version as determined
            by the algorithm software.
        revision : str
            %y%j%H%M%S formatted time when the file was created.
        extension : str
        """
        return cls._fmt.format(instrument=instrument,
                               level=level.value,
                               utc_start=utc_start.strftime(PRINTABLE_TS_FORMAT),
                               utc_end=utc_end.strftime(PRINTABLE_TS_FORMAT),
                               version=version,
                               revision=revision,
                               extension=extension)

    def _parse_filename_parts(self):
        """Parse the filename parts into objects from regex matched strings

        Returns
        -------
        : SimpleNamespace
            namespace object containing filename parts as parsed objects
        """
        d = self.regex_match(self.path)
        d['level'] = DataLevel(d['level'])
        d['utc_start'] = datetime.strptime(d['utc_start'], PRINTABLE_TS_FORMAT)
        d['utc_end'] = datetime.strptime(d['utc_end'], PRINTABLE_TS_FORMAT)
        return SimpleNamespace(**d)


class ManifestFilename(AbstractValidFilename):
    """Class for naming manifest files"""
    _regex = MANIFEST_FILE_REGEX
    _fmt = "libera_{manifest_type}_manifest_{created_time}.json"

    @classmethod
    def _format_filename_parts(cls,  # pylint: disable=arguments-differ
                               manifest_type: ManifestType,
                               created_time: datetime):
        """Construct a path from filename parts

        Parameters
        ----------
        manifest_type : ManifestType
            Input or output
        created_time : datetime
            Time of manifest creation (writing).
        """
        return cls._fmt.format(manifest_type=manifest_type.value.lower(),
                               created_time=created_time.strftime(PRINTABLE_TS_FORMAT))

    def _parse_filename_parts(self):
        """Parse the filename parts into objects from regex matched strings

        Returns
        -------
        : SimpleNamespace
            namespace object containing filename parts as parsed objects
        """
        d = self.regex_match(self.path)
        d['manifest_type'] = ManifestType(d['manifest_type'].upper())
        d['created_time'] = datetime.strptime(d['created_time'], PRINTABLE_TS_FORMAT)
        return SimpleNamespace(**d)


class EphemerisKernelFilename(AbstractValidFilename):
    """Class to construct, store, and manipulate an SPK filename"""
    _regex = SPK_REGEX
    _fmt = "libera_{spk_object}_{utc_start}_{utc_end}.bsp"

    @classmethod
    def _format_filename_parts(cls,  # pylint: disable=arguments-differ
                               spk_object: str,
                               utc_start: datetime,
                               utc_end: datetime):
        """Create an instance from a given path

        Parameters
        ----------
        spk_object : str
            Name of object whose ephemeris is represented in this SPK.
        utc_start : datetime
            Start time of data.
        utc_end : datetime
            End time of data.

        Returns
        -------
        : cls
        """
        return cls._fmt.format(spk_object=spk_object,
                               utc_start=utc_start.strftime(PRINTABLE_TS_FORMAT),
                               utc_end=utc_end.strftime(PRINTABLE_TS_FORMAT))

    def _parse_filename_parts(self):
        """Parse the filename parts into objects from regex matched strings

        Returns
        -------
        : SimpleNamespace
            namespace object containing filename parts as parsed objects
        """
        d = self.regex_match(self.path)
        d['utc_start'] = datetime.strptime(d['utc_start'], PRINTABLE_TS_FORMAT)
        d['utc_end'] = datetime.strptime(d['utc_end'], PRINTABLE_TS_FORMAT)
        return SimpleNamespace(**d)


class AttitudeKernelFilename(AbstractValidFilename):
    """Class to construct, store, and manipulate an SPK filename"""
    _regex = CK_REGEX
    _fmt = "libera_{ck_object}_{utc_start}_{utc_end}.bc"

    @classmethod
    def _format_filename_parts(cls,  # pylint: disable=arguments-differ
                               ck_object: str,
                               utc_start: datetime,
                               utc_end: datetime):
        """Create an instance from a given path

        Parameters
        ----------
        ck_object : str
            Name of object whose attitude is represented in this CK.
        utc_start : datetime
            Start time of data.
        utc_end : datetime
            End time of data.

        Returns
        -------
        : cls
        """
        return cls._fmt.format(ck_object=ck_object,
                               utc_start=utc_start.strftime(PRINTABLE_TS_FORMAT),
                               utc_end=utc_end.strftime(PRINTABLE_TS_FORMAT))

    def _parse_filename_parts(self):
        """Parse the filename parts into objects from regex matched strings

        Returns
        -------
        : SimpleNamespace
            namespace object containing filename parts as parsed objects
        """
        d = self.regex_match(self.path)
        d['utc_start'] = datetime.strptime(d['utc_start'], PRINTABLE_TS_FORMAT)
        d['utc_end'] = datetime.strptime(d['utc_end'], PRINTABLE_TS_FORMAT)
        return SimpleNamespace(**d)


def get_current_revision():
    """Get the current `r%y%j%H%M%S` string for naming file revisions.

    Returns
    -------
    : str
    """
    return f"r{datetime.utcnow().strftime('%y%j%H%M%S')}"


def format_version(semantic_version: str):
    """Formats a semantic version string X.Y.Z into a filename-compatible string like vMXmYpZ, for Major, minor, patch.

    Parameters
    ----------
    semantic_version : str
        String matching X.Y.Z where X, Y and Z are integers of any length

    Returns
    -------
    : str
    """
    major, minor, patch = semantic_version.split('.')
    return f"vM{major}m{minor}p{patch}"



"""Module containing CLI tool for creating SPICE kernels from packets"""
# Standard
import argparse
import warnings
from datetime import datetime
import logging
from pathlib import Path
import shutil
import subprocess  # nosec B404
import tempfile
# Installed
import numpy as np
import numpy.lib.recfunctions as nprf
from cloudpathlib import S3Path, AnyPath
from lasp_packets import parser, xtcedef
# Local
from libera_utils import spice_utils
from libera_utils.logutil import configure_task_logging
from libera_utils.config import config
from libera_utils.io import filenaming
from libera_utils import packets as libera_packets
from libera_utils import time
from libera_utils.io.smart_open import smart_open, smart_copy_file
from libera_utils.io.manifest import Manifest

logger = logging.getLogger(__name__)


def make_jpss_kernels_from_manifest(manifest_file_path: str or AnyPath,
                                    output_directory: str or AnyPath):
    """Alpha function triggering kernel generation from manifest file.

    If the manifest configuration field contains "start_time" and "end_time"
    fields then this function will select only packet data that falls in that
    range. If these are not given, then all packet data will be used.

    Parameters
    ----------
    manifest_file_path : str or AnyPath
        Path to the manifest file that includes end_time and start_time
        in the configuration section
    output_directory :  str or AnyPath
        Path to save the completed kernels
    Returns
    -------
    output_directory : str or AnyPath
        Path to the directory containing the completed kernels
    """
    # TODO: Consider cases to return/error if the entire range is not covered

    m = Manifest.from_file(manifest_file_path)
    m.validate_checksums()
    files_in_range = []

    if "start_time" not in m.configuration.keys():
        # No time range information is provided. Process all files in the manifest
        for file_entry in m.files:
            files_in_range.append(str(file_entry["filename"]))
    else:
        # Load desired time range from the manifest configuration
        start_time_text = m.configuration["start_time"]
        desired_start_time = datetime.strptime(start_time_text, '%Y-%m-%d:%H:%M:%S')
        end_time_text = m.configuration["end_time"]
        desired_end_time = datetime.strptime(end_time_text, '%Y-%m-%d:%H:%M:%S')

        # Load the packet files and check the time ranges against the manifest configuration
        # TODO update this if possible to use the metadata files when those are more defined

        for file_entry in m.files:
            file_path_from_list = file_entry["filename"]
            packet_data = get_spice_packet_data_from_filepaths([file_path_from_list])
            ephemeris_time = time.scs2e_wrapper(
                [f"{d}:{ms}:{us}" for d, ms, us in
                 zip(packet_data['ADAET1DAY'], packet_data['ADAET1MS'], packet_data['ADAET1US'])]
            )
            # Check if any of the packet data are in increasing order by comparing an array element to
            # its right neighbor and ensuring that is always greater or equal. If this is not true
            # throw an error
            if not np.all(ephemeris_time[:-1] <= ephemeris_time[1:]):
                raise ValueError(f"The data in {file_path_from_list} are not monotonic in time")
            packet_start_time = time.et_2_datetime(ephemeris_time[0])
            packet_end_time = time.et_2_datetime(ephemeris_time[-1])

            # Packet range starts before desired range - first packet or full data
            if packet_start_time < desired_start_time < packet_end_time:
                files_in_range.append(str(file_path_from_list))
            # Desired range starts before packet range - middle or end packet
            if desired_start_time < packet_start_time < desired_end_time:
                files_in_range.append(str(file_path_from_list))

        if not files_in_range:
            raise ValueError(f"No files contained packets in timerange ({desired_start_time}, {desired_end_time})")

    # Create the arguments to pass to the kernel generation
    parsed_args = argparse.Namespace(
        packet_data_filepaths=files_in_range,
        outdir=str(output_directory),
        overwrite=False,
        verbose=False
    )
    make_jpss_spk(parsed_args)
    make_jpss_ck(parsed_args)

    return output_directory


def get_spice_packet_data_from_filepaths(packet_data_filepaths):
    """Utility function to return an array of packet data from a list of file paths of raw JPSS APID 11
    geolocation packet data files.

     Parameters
    ----------
    packet_data_filepaths : list
        The list of file paths to the raw packet data

    Returns
    -------
    packet_data : ndarray
        The configured packet data. See packets.py for more details on structure
     """
    packet_definition_uri = AnyPath(config.get('JPSS_GEOLOCATION_PACKET_DEFINITION'))
    logger.info("Using packet definition %s", packet_definition_uri)

    with smart_open(packet_definition_uri) as packet_definition_filepath:
        packet_definition = xtcedef.XtcePacketDefinition(packet_definition_filepath)

    packet_parser = parser.PacketParser(packet_definition=packet_definition)

    packet_data = libera_packets.parse_packets(packet_parser, packet_data_filepaths)

    return packet_data

def make_jpss_spk(parsed_args: argparse.Namespace):
    """Create a JPSS SPK from APID 11 CCSDS packets.

    Parameters
    ----------
    parsed_args : argparse.Namespace
        Namespace of parsed CLI arguments

    Returns
    -------
    None
    """

    now = datetime.utcnow().strftime("%Y%m%dt%H%M%S")
    configure_task_logging(f'spk_generator_{now}',
                           app_package_name='libera_utils',
                           console_log_level=logging.DEBUG if parsed_args.verbose else None)

    logger.info("Starting SPK maker. This CLI tool creates an SPK from a list of geolocation packet files.")

    output_dir = AnyPath(parsed_args.outdir)
    logger.info("Writing resulting SPK to %s", output_dir)

    logger.info("Parsing packets...")
    packet_data = get_spice_packet_data_from_filepaths(parsed_args.packet_data_filepaths)
    logger.info("Done.")

    # Calculate and append a ET representation of the epochs. MKSPK is picky about time formats.
    ephemeris_time = time.scs2e_wrapper(
        [f"{d}:{ms}:{us}" for d, ms, us in
         zip(packet_data['ADAET1DAY'], packet_data['ADAET1MS'], packet_data['ADAET1US'])]
    )
    packet_data = nprf.append_fields(packet_data, 'ET', ephemeris_time, dtypes=(np.float64,))

    with tempfile.TemporaryDirectory(prefix='/tmp/') as tmp_dir:  # nosec B108
        tmp_path = Path(tmp_dir)
        spk_data_filepath = write_kernel_input_file(
            packet_data,
            filepath=tmp_path / 'mkspk_data.txt',
            fields=['ET', 'ADGPSPOSX', 'ADGPSPOSY', 'ADGPSPOSZ', 'ADGPSVELX', 'ADGPSVELY', 'ADGPSVELZ'])
        logger.info("MKSPK input data written to %s", spk_data_filepath)

        spk_setup_filepath = write_kernel_setup_file(
            config.get("MKSPK_SETUPFILE_CONTENTS"),
            filepath=tmp_path / 'mkspk_setup.txt')
        logger.info("MKSPK setup file written to %s", spk_setup_filepath)

        utc_start_str = time.et_2_datetime(ephemeris_time[0])
        utc_end_str = time.et_2_datetime(ephemeris_time[-1])
        spk_filename = filenaming.EphemerisKernelFilename.from_filename_parts(
            spk_object='jpss', utc_start=utc_start_str, utc_end=utc_end_str)
        output_filepath = tmp_path / spk_filename.path.name  # pylint: disable=no-member

        if parsed_args.overwrite is True:
            output_filepath.unlink(missing_ok=True)

        logger.info("Running MKSPK...")
        try:
            result = subprocess.run(['mkspk',  # nosec B603 B607
                                     '-setup', str(spk_setup_filepath),
                                     '-input', str(spk_data_filepath),
                                     '-output', str(output_filepath)],
                                    capture_output=True, check=True)
        except subprocess.CalledProcessError as cpe:
            logger.info("Captured stdout: \n%s", cpe.stdout.decode())
            if cpe.stderr:
                logger.error("Captured stderr: \n%s", cpe.stderr.decode())
            raise

        logger.info("Captured stdout:\n%s", result.stdout.decode())
        if result.stderr:
            logger.error(result.stderr.decode())
        logger.info("Finished! SPK written to %s", output_filepath)

        output_full_path = output_dir / spk_filename.path.name  # pylint: disable=no-member
        # Use smart copy here to avoiding using two nested smart_open calls
        # one call would be to open the newly created file, and one to open the desired location
        smart_copy_file(output_filepath, output_full_path)
        logger.info("SPK copied to %s", output_full_path)


def make_jpss_ck(parsed_args: argparse.Namespace):
    """Create a JPSS CK from APID 11 CCSDS packets.

    Parameters
    ----------
    parsed_args : argparse.Namespace
        Namespace of parsed CLI arguments

    Returns
    -------
    None
    """
    now = datetime.utcnow().strftime("%Y%m%dt%H%M%S")
    configure_task_logging(f'ck_generator_{now}',
                           app_package_name='libera_utils',
                           console_log_level=logging.DEBUG if parsed_args.verbose else None)

    logger.info("Starting CK maker. This CLI tool creates a CK from a list of geolocation packet files.")

    output_dir = AnyPath(parsed_args.outdir)
    logger.info("Writing resulting CK to %s", output_dir)

    logger.info("Parsing packets...")
    packet_data = get_spice_packet_data_from_filepaths(parsed_args.packet_data_filepaths)
    logger.info("Done.")

    # Add a column that is the SCLK string, formatted with delimiters, to the input data recarray
    attitude_sclk_string = [f"{row['ADAET2DAY']}:{row['ADAET2MS']}:{row['ADAET2US']}" for row in packet_data]
    packet_data = nprf.append_fields(packet_data, 'ATTSCLKSTR', attitude_sclk_string)

    with tempfile.TemporaryDirectory(prefix='/tmp/') as tmp_dir:  # nosec B108
        tmp_path = Path(tmp_dir)
        ck_data_filepath = write_kernel_input_file(
            packet_data,
            filepath=tmp_path / 'msopck_data.txt',
            fields=['ATTSCLKSTR', 'ADCFAQ4', 'ADCFAQ1', 'ADCFAQ2', 'ADCFAQ3'],
            fmt=['%s', '%.16f', '%.16f', '%.16f', '%.16f']
        )  # produces w + i + j + k in SPICE_QUATERNION style
        logger.info("MSOPCK input data written to %s", ck_data_filepath)

        ck_setup_filepath = write_kernel_setup_file(
            config.get("MSOPCK_SETUPFILE_CONTENTS"),
            filepath=tmp_path / 'msopck_setup.txt')
        logger.info("MSOPCK setup file written to %s", ck_setup_filepath)

        utc_start_str = time.et_2_datetime(time.scs2e_wrapper(attitude_sclk_string[0]))
        utc_end_str = time.et_2_datetime(time.scs2e_wrapper(attitude_sclk_string[-1]))
        ck_filename = filenaming.AttitudeKernelFilename.from_filename_parts(
            ck_object='jpss', utc_start=utc_start_str, utc_end=utc_end_str)
        output_filepath = tmp_path / ck_filename.path.name  # pylint: disable=no-member

        if parsed_args.overwrite is True:
            output_filepath.unlink(missing_ok=True)

        logger.info("Running MSOPCK...")
        try:
            result = subprocess.run(['msopck',  # nosec B603 B607
                                     str(ck_setup_filepath), str(ck_data_filepath), str(output_filepath)],
                                    capture_output=True, check=True)
        except subprocess.CalledProcessError as cpe:
            logger.info("Captured stdout: \n%s", cpe.stdout.decode())
            if cpe.stderr:
                logger.error("Captured stderr: \n%s", cpe.stderr.decode())
            raise

        logger.info("Captured stdout:\n%s", result.stdout.decode())
        if result.stderr:
            logger.error(result.stderr.decode())
        logger.info("Finished! CK written to %s", output_filepath)

        output_full_path = output_dir / ck_filename.path.name  # pylint: disable=no-member
        # Use smart copy here to avoiding using two nested smart_open calls
        # one call would be to open the newly created file, and one to open the desired location
        smart_copy_file(output_filepath, output_full_path)
        logger.info("CK copied to %s", output_full_path)

def make_azel_ck(parsed_args: argparse.Namespace):
    """Create a Libera Az-El CK from CCSDS packets

    Parameters
    ----------
    parsed_args : argparse.Namespace
        Namespace of parsed CLI arguments

    Returns
    -------
    None
    """
    print(parsed_args)
    raise NotImplementedError("CK generation for the Az-El mechanism isn't implemented yet.")


def write_kernel_input_file(data: np.ndarray, filepath: str or Path or S3Path,
                            fields: list = None, fmt: str or list = "%.16f"):
    """Write ephemeris and attitude data to MKSPK and MSOPCK input data files, respectively.

    See MSOPCK documentation here:
        https://naif.jpl.nasa.gov/pub/naif/toolkit_docs/C/ug/msopck.html
    See MKSPK documentation here:
        https://naif.jpl.nasa.gov/pub/naif/toolkit_docs/C/ug/mkspk.html

    Parameters
    ----------
    data : np.ndarray
        Structured array (named, with data types) of attitude or ephemeris data.
    filepath : str or Path
        Filepath to write to.
    fields : list
        Optional. List of field names to write out to the data file. If not specified, assume fields are already
        in the proper order.
    fmt : str or list
        Format specifier(s) to pass to np.savetxt. Default is to assume everything should be floats with 16 decimal
        places of precision (%.16f). If a list is passed, it must contain a format specifier for each column in data.

    Returns
    -------
    : Path
        Absolute path to written file.
    """
    if fields:
        np.savetxt(filepath, data[fields], delimiter=" ", fmt=fmt)
    else:
        np.savetxt(filepath, data[:], delimiter=" ", fmt=fmt)
    return filepath.absolute()


def write_kernel_setup_file(data: dict, filepath: Path):
    """Write an MSOPCK or MKSPK compatible setup file of key-value pairs.
    See documentation here: https://naif.jpl.nasa.gov/pub/naif/toolkit_docs/C/ug/msopck.html#Input%20Data%20Format

    Parameters
    ----------
    data : dict
        Dictionary of key-value pairs to write to the setup file.
    filepath : Path
        Filepath to write to.

    Returns
    -------
    : Path
        Absolute path to written file.
    """
    with open(filepath, 'x+', encoding='utf_8') as fh:
        fh.write("\\begindata\n")
        for key, value in data.items():
            if key in ('PATH_VALUES', 'PATH_SYMBOLS', 'KERNELS_TO_LOAD'):
                inside = ", ".join([f"\n\t'{item}'" for item in value])
                value_str = f"({inside}\n)"
            elif key in ('LSK_FILE_NAME', 'LEAPSECONDS_FILE'):
                lsk_url = spice_utils.find_most_recent_naif_kernel(spice_utils.NAIF_LSK_INDEX_URL,
                                                                   spice_utils.NAIF_LSK_REGEX)
                lsk = spice_utils.KernelFileCache(lsk_url)
                if len(str(lsk.kernel_path)) > 78:
                    # MSOPCK is limited to 80 character file names (including single quotes)
                    # so we copy the kernel to the same directory where we are writing this setup file
                    copied_kernel = shutil.copy(str(lsk.kernel_path), filepath.parent)
                    value_str = f"'{copied_kernel}'"
                else:
                    value_str = f"'{str(lsk.kernel_path)}'"
            elif key == 'SCLK_FILE_NAME' and len(value) > 78:
                # MSOPCK is limited to 80 character file names (including single quotes) so we copy the SCLK to the
                # same directory where we are writing this setup file
                copied_sclk = shutil.copy(value, filepath.parent)
                value_str = f"'{copied_sclk}'"
            elif isinstance(value, str):
                value_str = f"'{value}'"
            elif isinstance(value, list):
                list_str = " ".join(value)
                value_str = f"'{list_str}'"
            elif isinstance(value, dict):
                dict_str = " ".join([f"\n\t'{k}={v}'" for k, v in value.items()])
                value_str = f"({dict_str}\n)"
            else:
                value_str = f"{value}"
            if len(value_str) > 80:
                warnings.warn("Detected a SPICE setup file value that is over 80 characters. "
                              f"This will likely cause an error. {value_str}")
            fh.write(f"{key}={value_str}\n")
        fh.write("\\begintext\n")
        fh.seek(0)
        logger.info("Setup file contents:\n%s", ''.join(fh.readlines()))
    return filepath.absolute()



"""Module for reading packet data"""
# Standard
import logging
# Installed
import bitstring
from lasp_packets import parser
import numpy as np
import numpy.lib.recfunctions as nprf
# Local
from libera_utils.io.smart_open import smart_open

logger = logging.getLogger(__name__)


def array_from_packets(packets: list, apid: int = None):
    """Create an array from a packet generator, as returned from a lasp_packets PacketParser. This function assumes
    that the fields and format for every packet is identical for a given APID.

    Parameters
    ----------
    packets : list
        List of lasp_packets.parser.Packet objects.
    apid : int
        Application Packet ID to create an array from. We can only create an array for a single APID because we need
        to assume the same fields in every packet. If not specified, every packet must be of the same APID.

    Returns
    -------
    : np.recarray
        Record array with one column per field name in the packet type. Values are derived if a derived value exists,
        otherwise, the values are the raw values.
    """
    apids_present = {packet.header['PKT_APID'].raw_value for packet in packets}
    if apid is not None and apid not in apids_present:
        raise ValueError("Requested APID not found in parsed packets.")
    if apid is None and len(apids_present) > 1:
        raise ValueError("Multiple APIDs present. To create an array you must specify which APID you want.")

    apid = apid or apids_present.pop()

    field_values = [
        tuple(pdi.derived_value or pdi.raw_value for pdi in packet.data.values())
        for packet in packets
        if packet.header['PKT_APID'].raw_value == apid
    ]
    names = tuple(pdi.name for pdi in packets[0].data.values())  # Get data field names from the first Packet
    formats = tuple(type(val) if not isinstance(val, str) else object for val in field_values[0])
    return np.array(field_values, dtype={'names': names, 'formats': formats})


def parse_packets(packet_parser: parser.PacketParser, packet_data_filepaths: list, apid: int = None):
    """Parse a recarray from a list of packet filepaths, assuming the same parser for all

    Parameters
    ----------
    packet_parser : parser.PacketParser
        Parser, already initialized with the anticipated definition.
    packet_data_filepaths : list
        List of filepaths to packets files.
    apid : int
        Filter on APID so we don't get mismatches in case the parser finds multiple parsable packet definitions
        in the files. This can happen if the XTCE document contains definitions for multiple packet types and >1 of
        those packet types is present in the packet data files.

    Returns
    -------
    : np.recarray
        Concatenated arrays of packet data.
    """
    data_arrays = []
    for packet_data_filepath in packet_data_filepaths:
        logger.info("Packet data filepath %s", packet_data_filepath)
        with smart_open(packet_data_filepath) as packet_bytes:
            binary_packet_data = bitstring.ConstBitStream(bytes=packet_bytes.read())

        packet_generator = packet_parser.generator(binary_packet_data)
        data_arrays.append(array_from_packets(list(packet_generator), apid=apid))

    packet_data = nprf.stack_arrays(data_arrays, asrecarray=True, usemask=False)  # Stack recarrays
    # Remove any duplicates in case we got the same packet twice in different files
    return np.unique(packet_data, axis=0)



"""db module"""
# Standard
import re
# Installed
from sqlalchemy.ext.declarative import declarative_base, declared_attr
# Local
from .database import _DatabaseManager


def camel_to_snake(name):
    """Convert a CamelCase string to a snake_case.

    Parameters
    ----------
    name: str
        String to convert.

    Returns
    -------
    : str
    """
    name = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', name)
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', name).lower()


class BaseAugmentation:
    """
    Augmentation class that provides some standard behavior to all of our models.
    See https://docs.sqlalchemy.org/en/13/orm/extensions/declarative/mixins.html#augmenting-the-base
    """
    __name__ = None  # Gets overridden later. We just want our class declaration to know this exists.
    __abstract__ = True
    __table_args__ = {"schema": "sdp"}

    @declared_attr
    def __tablename__(cls):  # pylint: disable=E0213
        """Set the __tablename__ of every model to be the snake_case form of the class name."""
        return camel_to_snake(cls.__name__)


# TODO: This will need to change when upgrading to SQLAlchemy 2.0
Base = declarative_base(cls=BaseAugmentation)


# Convenience method for getting database managers
getdb = _DatabaseManager.get
"""Convenience alias for :meth:`libera_utils.db.database._DatabaseManager.get`"""
